import torch

class ResBlock(torch.nn.Module):
    def __init__(self, units_in, units_out, kernel_size=3, pad_right=True, dilation=1, dropout=0):
        super(ResBlock, self).__init__()
        self.conv1 = torch.nn.Conv1d(units_in, units_out*2, kernel_size, dilation=dilation)
        self.conv2 = torch.nn.Conv1d(units_out, units_out*2, kernel_size, dilation=dilation)
        self.drop = torch.nn.Dropout(dropout)
        padding = (kernel_size*dilation) -dilation
        padding_left = padding // 2
        padding_right = padding // 2
        rem = padding % 2
        if pad_right:
            padding_right = padding_right + rem
        else:
            padding_left = padding_left + rem
        self.pad1 = torch.nn.ConstantPad1d((int(padding_left), int(padding_right)), 0)
        self.resizing = torch.nn.Conv1d(units_in, units_out, 1) if units_in != units_out else None


    def forward(self, x):
        #input: (N, C, L)
        res = x if self.resizing is None else self.resizing(x)
        out = self.conv1(self.pad1(x))
        out = torch.nn.functional.glu(out, dim=1)
        out = self.drop(out)
        out = self.conv2(self.pad1(out))
        out = torch.nn.functional.glu(out, dim=1)
        out = self.drop(out)
        out = out + res
        return out


class CharInputBlock(torch.nn.Module):
    def __init__(self, charnr, embsize, hidden_units, kernel_size=3, max=True, dilation=1, kernel_sizes=None, dilations=None, dropout=0):
        super(CharInputBlock, self).__init__()
        self.embedding = torch.nn.Embedding(charnr, embsize)
        if dilations is None:
            dilations = [dilation]*len(hidden_units)
        if kernel_sizes is None:
            kernel_sizes = [kernel_size]*len(hidden_units)
        layers = []
        layer = ResBlock(embsize, hidden_units[0], kernel_sizes[0], dilation=dilations[0], dropout=dropout)
        layers.append(layer)

        for i in range(1, len(hidden_units)):
            layer = ResBlock(hidden_units[i-1], hidden_units[i], kernel_sizes[i], dilation=dilations[i], dropout=dropout)
            layers.append(layer)
        self.layers = torch.nn.Sequential(*layers)
        if max:
            self.pool = torch.nn.AdaptiveMaxPool1d(1)
        else:
            self.pool = torch.nn.AdaptiveAvgPool1d(1)


    def forward(self, x):
        embs = self.embedding(x).transpose(1, 2)
        out = self.layers(embs.contiguous())
        out = self.pool(out)
        return out


class TestNet(torch.nn.Module):
    def __init__(self, charvocabsize, embsize, hidden_units_char, hidden_units_post, kernel_size=3, pad_right=True, dilation=1, kernel_sizes_char=None, kernel_sizes_post=None, dilations_char=None, dilations_post=None, dropout=0):
        super(TestNet, self).__init__()
        self.inputlayer = CharInputBlock(charvocabsize, embsize, hidden_units_char, dilation=dilation, kernel_size=kernel_size, kernel_sizes=kernel_sizes_char, dilations=dilations_char, dropout=dropout)
        if dilations_post is None:
            dilations_post = [dilation]*len(hidden_units_post)
        if kernel_sizes_post is None:
            kernel_sizes_post = [kernel_size]*len(hidden_units_post)
        postlayers = [ResBlock(hidden_units_char[-1], hidden_units_post[0], kernel_sizes_post[0], dilation=dilations_post[0], dropout=dropout)]
        for i in range(1, len(hidden_units_post)):
            layer = ResBlock(hidden_units_post[i-1], hidden_units_post[i], kernel_sizes_post[i], dilation=dilations_post[i], dropout=dropout)
            postlayers.append(layer)
        self.postlayers = torch.nn.Sequential(*postlayers)
        padding = kernel_size -1
        padding_left = padding // 2
        padding_right = padding // 2
        rem = padding % 2
        if pad_right:
            padding_right = padding_right + rem
        else:
            padding_left = padding_left + rem
        self.pad = torch.nn.ConstantPad1d((int(padding_left), int(padding_right)), 0)
        self.out_conv = torch.nn.Conv1d(hidden_units_post[-1], 17, kernel_size)
        self.out_log = torch.nn.LogSoftmax(dim=1)


    def forward(self, sentence):
        ns = []
        for w in sentence:
            ns.append(self.inputlayer(w))
        ns = torch.cat(ns, 2)
        ns = ns.contiguous()
        x = self.postlayers(ns)
        x = self.pad(x)
        x = self.out_conv(x)
        x = self.out_log(x)
        return x
