import torch
import conlluparser
import utils
import argparse
import model_relu
import model_single
import model_double
import time


parser = argparse.ArgumentParser(description="Test runner for POST")
parser.add_argument("--dilation", "-d", type=int, default=1)
parser.add_argument("--kernel_size", "-ks", type=int, default=3)
# The arguments above are meant as shortcuts and defines a single value for all layers.
# The values given for dilation and kernel size with the arguments below will override those given above.
parser.add_argument("--dilation_post", "-dpost", type=int, nargs='*')
parser.add_argument("--dilation_char", "-dchar", type=int, nargs='*')
parser.add_argument("--kernel_size_post", "-kspost", type=int, nargs='*')
parser.add_argument("--kernel_size_char", "-kschar", type=int, nargs='*')
parser.add_argument("--hidden_units_char", "-huchar", type=int, nargs='*', default=[200])
parser.add_argument("--hidden_units_post", "-hupost", type=int, nargs='*', default=[200])
parser.add_argument("--embsize", "-es", type=int, default=300)
parser.add_argument("--filename_train", "-ftrain", required=True)
parser.add_argument("--filename_test", "-ftest", required=True)
parser.add_argument("--charvocab", "-cv")
parser.add_argument("--optimizer", "-opt", default="Adagrad", choices=["Adam", "Adagrad", "SGD"])
parser.add_argument("--pos_tags", "-post", required=True)
parser.add_argument("--batchtype", "-bt", choices=["sentence", "batch"], default="batch")
parser.add_argument("--batchlength", "-bl", default=80, type=int)
parser.add_argument("--learning_rate", "-lr", type=float, default=0.01)
parser.add_argument("--momentum", "-mom", type=float, default=0)
parser.add_argument("--dropout", "-do", type=float, default=0)
parser.add_argument("--training_rounds", "-tr", type=int, default=10)
parser.add_argument("--disable_cuda", action="store_true")
parser.add_argument("--nettype", "-nt", choices=["glu1", "glu2", "relu"], default="glu1")

vars = parser.parse_args()

if not vars.disable_cuda and torch.cuda.is_available():
    device = torch.device("cuda:1")
    cuda = True
else:
    device = torch.device("cpu")
    cuda = False

#This sets the seed for torch
torch.manual_seed(597394801387)

train_data = conlluparser.readconllufile(vars.filename_train)
train_data = utils.prepareforpostagging(train_data)

test_data = conlluparser.readconllufile(vars.filename_test)
test_data = utils.prepareforpostagging(test_data)

if vars.charvocab is not None:
    chardict = utils.readchardict(vars.charvocab)
else:
    chardict = {"OOV": 1}
    for s in train_data:
        for w in s:
            for c in w[0]:
                if c not in chardict:
                    chardict[c] = len(chardict) + 1

tags2int, int2tags = utils.readpostfile(vars.pos_tags)
labels = []
for i in range(0, len(int2tags)):
    labels.append(int2tags[i])

if vars.batchtype == "sentence":
    train_words, train_tags = utils.getsentencebatch(train_data, chardict, tags2int, vars.batchlength, device=device)
    test_words, test_tags = utils.getsentencebatch(test_data, chardict, tags2int, vars.batchlength, device=device)
else:
    train_words, train_tags = utils.getsamelengthbatches(train_data, chardict, tags2int, vars.batchlength, device=device)
    test_words, test_tags = utils.getsamelengthbatches(test_data, chardict, tags2int, vars.batchlength, device=device)

#Here the seed is set to make the batches the same order
train_words, train_tags = utils.shuffle_batches(train_words, train_tags, seed=95737413863)

if vars.nettype == "glu1":
    net = model_single.TestNet(len(chardict)+1, vars.embsize, vars.hidden_units_char, vars.hidden_units_post, kernel_size=vars.kernel_size, dilation=vars.dilation, kernel_sizes_char=vars.kernel_size_char, kernel_sizes_post=vars.kernel_size_post, dilations_char=vars.dilation_char, dilations_post=vars.dilation_post, dropout=vars.dropout)
elif vars.nettype == "glu2":
    net = model_double.TestNet(len(chardict)+1, vars.embsize, vars.hidden_units_char, vars.hidden_units_post, kernel_size=vars.kernel_size, dilation=vars.dilation, kernel_sizes_char=vars.kernel_size_char, kernel_sizes_post=vars.kernel_size_post, dilations_char=vars.dilation_char, dilations_post=vars.dilation_post, dropout=vars.dropout)
elif vars.nettype == "relu":
    net = model_relu.TestNet(len(chardict)+1, vars.embsize, vars.hidden_units_char, vars.hidden_units_post, kernel_size=vars.kernel_size, dilation=vars.dilation, kernel_sizes_char=vars.kernel_size_char, kernel_sizes_post=vars.kernel_size_post, dilations_char=vars.dilation_char, dilations_post=vars.dilation_post, dropout=vars.dropout)

net.to(device)

lossfunc = torch.nn.NLLLoss()

if vars.optimizer.lower() == "adam":
    optimizer = torch.optim.Adam(net.parameters(), lr=vars.learning_rate)
elif vars.optimizer.lower() == "adagrad":
    optimizer = torch.optim.Adagrad(net.parameters(), lr=vars.learning_rate)
elif vars.optimizer.lower() == "sgd":
    optimizer = torch.optim.SGD(net.parameters(), vars.learning_rate, momentum=vars.momentum)

acc_loss = 0
print("Parameter count:")
npars = sum(p.numel() for p in net.parameters() if p.requires_grad)
print(npars)

with open("trainresults.txt", "a") as f:
    s = "Variables:\n"
    s += str(vars)
    s += "\nParameters:\n"
    s += str(npars)
    f.write(s)

trainingstats = utils.StatObject(labels)
teststats = utils.StatObject(labels)
t1 = time.time()

for j in range(0, vars.training_rounds):
    print("Epoch %d of %d" % (j+1, vars.training_rounds))
    #Train one epoch
    trainingstats.clear()
    teststats.clear()
    net.train()
    for i in range(0, len(train_words)):
        optimizer.zero_grad()
        out = net(train_words[i])
        out = out.contiguous()
        loss = lossfunc(out, train_tags[i])
        loss.backward()
        optimizer.step()
        acc_loss += loss.item()
        outputs = torch.argmax(out, dim=1)
        for k in range(0, train_tags[i].size(1)):
            ti = train_tags[i][0][k].item()
            oi = outputs[0][k].item()
            trainingstats.add_prediction(ti, oi)
        if (i+1)%100 == 0:
            print("Batch %5d" % (i+1))
            print("Loss: %.3f" % acc_loss)
            with open("accloss.txt", "a") as f:
                s = "%d, %d: %5.3f\n" % (j, i, acc_loss)
                f.write(s)
            acc_loss = 0
    trainingstats.evaluate()
    print("Training accuracy: %.3f" % trainingstats.accuracy)
    with open("trainresults.txt", "a") as f:
        s = "\nEpoch %d\n" % j
        s += str(trainingstats)
        f.write(s)

    #Test at end of epoch
    net = net.eval()
    for i in range(0, len(test_words)):
        out = net(test_words[i])
        out = out.contiguous()
        outputs = torch.argmax(out, dim=1)
        for k in range(0, test_tags[i].size(1)):
            ti = test_tags[i][0][k].item()
            oi = outputs[0][k].item()
            teststats.add_prediction(ti, oi)
    teststats.evaluate()
    print("Test accuracy: %.3f" % teststats.accuracy)
    with open("testresults.txt", "a") as f:
        s = "\nEpoch %d\n" % j
        s += str(teststats)
        f.write(s)

t2 = time.time()
print("Total time:")
tottime = t2-t1
print(tottime)
with open("trainresults.txt", "a") as f:
    s = "\nTotal time: %.2f\n" % tottime
    f.write(s)
print(teststats.accuracy)
